"""
Contains astronomical and physical constants for use in SunPy or other
places.

"""

from __future__ import absolute_import

from sunpy.sun.sun import *

from . import _constants
from . import constants
