# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2017-11-16 12:04:31
from __future__ import unicode_literals
import datetime

version = "2.0.2"
githash = "d23a53f46dd1c3703e5eee63dca3f53bd18a4e8b"


major = 2
minor = 0
bugfix = 2

release = True
timestamp = datetime.datetime(2017, 11, 16, 12, 4, 31)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
